# Some example local_scan() functions that could just as easily be
# implemented with Exim ACLs.  You probably wouldn't want to use any of
# these in real life, on the assumption that using ACLs is more
# efficient than calling Python code.  But there's only so far you can
# go with ACLs -- once you want to get out of that little box, the
# easiest step up is to write a local_scan() in Python, possibly
# modelled on one of these.

__revision__ = "$Id: acl_equiv.py,v 1.4 2002/10/20 21:46:23 gward Exp $"

import re
from elspy import RejectMessage, TempRejectMessage


# Reject messages with 8-bit characters in the "Subject" or "From"
# headers.  This is a hallmark of Asian spammers, and it violates RFC
# 2822.  Note that we require three consecutive 8-bit characters -- the
# idea is that we let Europeans with misconfigured mail clients get away
# with a little bit of non-compliance.  Change the regex if you want to
# be more strict.  Equivalent to the ACL statements:
#   deny message   = 8-bit characters not allowed in subject\n\
#                    (see RFC 2822, sections 3.6.5, 2.2.1)
#        condition = ${if match {$h_subject:}{\N[\x80-\xFF]{3,}\N} {1}}
#
#   deny message   = 8-bit characters not allowed in header addresses\n\
#                    (see RFC 2822, sections 3.4, 3.2.4, 3.2.5)
#        condition = ${if match {$h_from:}{\N[\x80-\xFF]{3,}\N} {1}}

eightbit_re = re.compile(r'[\x80-\xFF]{3,}')

def local_scan_8bitheaders (info, headers, fd):
    subject = headers.get("subject", "")
    if eightbit_re.search(subject):
        raise RejectMessage(
            "8-bit characters not allowed in subject\n"
            "(see RFC 2822, sections 3.6.5, 2.2.1)")

    from_hdr = headers.get("from", "")
    if eightbit_re.search(from_hdr):
        raise RejectMessage(
            "8-bit characters not allowed in header addresses\n"
            "(see RFC 2822, sections 3.4, 3.2.4, 3.2.5)")


# Reject messages with encoded subject, where the encoding charset is
# one that I don't like because all the mail I ever get in that charset
# is spam.  If you get a lot of Chinese or Korean email, this obviously
# won't work for you.  Roughly equivalent to the ACL statement:
#  deny condition = ${if match {$h_subject:}{\N=\?(big5|euc-kr)\?[QB]\?\N} {1}}
#       message   = banned charset in subject

BANNED_CHARSETS = ("big5", "euc-kr")

encoded_re = re.compile(r'=\?([\w\-]+)\?[QB]\?')

def local_scan_charset (info, headers, fd):
    subject = headers.get("subject", "").lower()
    match = encoded_re.search(subject)
    if match:
        charset = match.group(1)
        if charset in BANNED_CHARSETS:
            raise RejectMessage("banned charset in subject")
