# elspy example -- reject messages with bogus or unbelievable "Date" header
#
# This code is based on a strict interpretation of RFC 2822 section 3.3;
# the idea is to reject *any* sort of bogosity in the "Date:" header:
#   
#   * invalid syntax, eg. any of the following would be rejected:
#       "Mon 1 Jul 2002 10:11:41 -0400"  (missing comma)
#       "Mon,  1 Jul 2002 10:11:41"      (missing timezone)
#       "Fri, 22 Feb 2002 09:31:14 EST"  (obsolete timezone syntax)
#       "Sat, 18 Dec 99 15:23:35 EST"    (obsolete year syntax)
#       "Sunday, November 07, 1999 10:58 PM"  (bad day, bad month, no seconds)
#
#   * out-of-range hour/minute/second
#
#   * out-of-range timezone
#
#   * inconsistent day-of-week, eg. "Tue,  1 Jul 2002 10:11:41"
#     (1 July 2002 was a Monday!)
#
#   * too far behind the current date; this is configurable by
#     editing MESSAGE_TOO_OLD below
#
#   * too far ahead of the current date; configurable by editing
#     MESSAGE_TOO_NEW below
# 

# Implementation notes.
#
# The date syntax recognized is from the grammar in RFC 2822 section
# 3.3, with "obsolete syntax" removed -- ie. we recognize precisely two
# date formats:
#   Mon,  1 Jul 2002 10:11:41 -0400
#   1 Jul 2002 10:11:41 -0400
#
# I didn't use the standard rfc822.parse_date() or parse_date_tz()
# functions for a couple of reasons:
#   * I had already written most of this code before I
#     remembered about them (d'ohh!)
#   * rfc822's date parsing is not as strict as what I implemented
#     here, and it would be hard to generate useful error messages
#     if I used a blackbox parser

# Copyright (c) 2002 Gregory P. Ward.  All rights reserved.
# See README.txt for licensing information.
#
# The elspy home page is at http://elspy.sourceforge.net/ .

__revision__ = "$Id: bogus_date.py,v 1.5 2002/10/21 22:20:14 gward Exp $"

import re
from time import time, localtime, gmtime, mktime, asctime
from elspy import RejectMessage

# If the "Date" header is behind the current time by more than
# MESSAGE_TOO_OLD hours (default: 7 days), reject the message.
MESSAGE_TOO_OLD = 7*24

# If the "Date" header is ahead of the current time by more than
# MESSAGE_TOO_NEW hours (default: 24 hours), reject the message.
MESSAGE_TOO_NEW = 24


# Regex derived from grammar in RFC 2822 section 3.3.  Does not
# recognize obsolete syntax.
date_re = re.compile(r'\s*'             # leading whitespace
                     r'(?: (\w{3}) ,)?' # day-of-week (optional)
                     r'\s* (\d\d?)'     # day of month (1 or 2 digits)
                     r'\s+ (\w{3}) \s+' # month name
                     r'(\d{4})'         # year (4 digits)
                     r'\s+'             # whitespace separating date and time
                     r'(\d\d):(\d\d):(\d\d)'# hh:mm:ss
                     r'\s+'
                     r'([\+\-])(\d\d)(\d\d)',  # timezone (+hhmm or -hhmm)
                     re.VERBOSE)

# English only -- that's what the RFC specifies!
WEEKDAYS = ["Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun"]
MONTHS = [None, "Jan", "Feb", "Mar", "Apr", "May", "Jun",
                "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"]

WEEKDAY_VALUE = {}
for i in range(len(WEEKDAYS)):
    WEEKDAY_VALUE[WEEKDAYS[i]] = i

MONTH_VALUE = {}
for i in range(1, len(MONTHS)):
    MONTH_VALUE[MONTHS[i]] = i


class ParseError (Exception):
    pass

def parse_date (date):
    match = date_re.match(date)
    if not match:
        raise ParseError("invalid syntax")

    # Convert everything to integers in order to do date calculations.
    (weekday_name, monthday, month_name,
     year, hour, min, sec,
     tzdir, tzhour, tzmin) = match.groups()
    if not MONTH_VALUE.has_key(month_name):
        raise ParseError("invalid month name: %r" % month_name)

    month = MONTH_VALUE[month_name]
    (monthday, year, hour, min, sec, tzhour, tzmin) = \
        map(int, (monthday, year, hour, min, sec, tzhour, tzmin))

    # Sanity check hour/min/sec (sec <= 60 because of leap seconds!).
    if hour > 23 or min > 59 or sec > 60:
        raise ParseError("invalid time: %02d:%02d:%02d" % (hour, min, sec))

    if tzmin > 59:
        raise ParseError("invalid timezone: %s%02d%02d"
                         % (tzdir, tzhour, tzmin))

    # Timezone in seconds relative to UTC.  Positive value means ahead
    # of UTC (eg. continental Europe), negative value means behind UTC
    # (eg. North America).
    tz = (tzhour*3600) + (tzmin*60)
    if tzdir == "-":
        tz = -tz

    # Sanity-check day-of-week, if it was supplied (it's optional
    # according to the grammar in RFC 2822).  Need to send it through
    # mktime() and localtime() to do this, which requires lying and
    # pretending that we have a local time, but I don't think that's a
    # problem.
    if weekday_name:
        if not WEEKDAY_VALUE.has_key(weekday_name):
            raise ParseError("invalid weekday: %r" % weekday_name)
        weekday = WEEKDAY_VALUE[weekday_name]

        t1 = (year, month, monthday, hour, min, sec, -1, -1, -1)
        t2 = localtime(mktime(t1))
        if t2[6] != weekday:
            raise ParseError("inconsistent weekday: expected %r, got %r"
                             % (WEEKDAYS[t2[6]], weekday_name))

    return (year, month, monthday, weekday, hour, min, sec, tz)


def local_scan (info, headers, fd):
    # Exim always adds a date header if one is missing, which means
    # that we can't reject messages for not having one.  But it
    # also means we can charge ahead here without having to worry
    # about 'date' being None.
    date = headers.get("Date")

    # Parse the date; reject the message if we couldn't parse it or
    # its internally inconsistent.
    try:
        (year, month, monthday, weekday, hour, min, sec, remote_tz) = \
            parse_date(date)
    except ParseError, err:
        raise RejectMessage(
            "bad date header: %s\n"
            "(see RFC 2822, section 3.3)" % err)

    # Now we need to compute the difference between the date parsed from
    # the message and the current date.  This is tricky because of
    # timezones, and the fact that Python's time module doesn't expose
    # useful timezone information (even though libc makes it available
    # on some platforms).  The fact that mktime() only works on local
    # times, and there is no UTC equivalent, doesn't help.  So here's my
    # hackish method:
    #   - determine local timezone by computing the difference
    #     between mktime() on a local time tuple and a UTC time
    #     tuple
    #   - determine remote timezone (ie. the timezone encoded in
    #     the "Date" header just parsed) by simple arithmetic on
    #     'zone' (eg. +0200 = two hours east [ahead] of UTC = 7200 sec)
    #   - adjust the remote time to UTC with mktime() and the
    #     difference between the local and remote timezones
    #
    # Example: say someone in Paris sends a message at
    # 2002-06-24 03:30 +0200 (ie. two hours ahead of UTC, because
    # they're on DST).  Five minutes later, the message lands on
    # my computer in Montreal, where the time is 2002-06-23 21:35 -0400
    # (four hours behind UTC).  The "Date" header on that message
    # should look like this:
    #   Date: Mon, 24 Jun 2002 03:30:00 +0200
    #
    # Using localtime(time()), we get this time tuple for the current
    # local time in Montreal:
    #   (2002, 6, 23, 21, 35, 0, 6, 174, 1)
    # (the last three numbers mean this is the 6th day of the week,
    # the 174th day of the year, and we're in daylight savings time).
    # gmtime(time()) gives the time tuple for the current time in UTC:
    #   (2002, 6, 24, 1, 35, 0, 0, 175, 0)
    # If we replace the is_dst flag here with is_dst from the local
    # time tuple, we can pass both tuples to mktime() and compute the
    # difference to get the local time zone: -14400, meaning
    # 14400 seconds behind (west) of UTC, or four hours.  This
    # appears to be the only way to figure out the current timezone
    # in Python -- ughh.
    #
    # Parsing the "Date" header and stealing the is_dst flag from
    # the local time tuple (which assumes that Paris is on DST
    # whenever Montreal is, which isn't true -- this will lead to
    # a one-hour error for a week or so each October and April, sigh),
    # we construct this time tuple for the remote time:
    #   (2002, 6, 24, 3, 30, -1, -1, 1)
    # (the -1's are because we don't care about the day-of-week or
    # day-of-year, and mktime() ignores those anyways).
    #
    # We then pass this to mktime(), which assumes the local timezone,
    # to get a bogus seconds-since-epoch value -- it will be ahead by
    # six hours, the difference between the remote and local timezones.
    # Since we have figured out the local timezone, and parsed the
    # remote timezone out of the "Date" header, it's easy to adjust
    # this bogus time value to get the correct time that the "Date"
    # header represents, as seconds since epoch UTC.
    # 
    # Given that, it's a simple subtraction to compute the difference
    # between the "Date" header and the current time, and reject the
    # message if the "Date" header is wildly off.  The definition of
    # "wildly off" needs to be pretty generous, given the known error in
    # my algorithm and the vagaries of Internet email transport.

    # Compute current timezone.  Can't use time.timezone because it
    # doesn't seem to take daylight savings time into account.  ;-(
    now = time()
    local_time = localtime(now)
    utc_munged = gmtime(now)[0:-1] + (local_time[-1],)
    local_tz = mktime(local_time) - mktime(utc_munged)

    # Construct remote time tuple -- this is where the one-hour error
    # for a week or so in October and April creeps in (assuming the two
    # timezones are in different DST regimes -- eg. within North America
    # or within Europe, this won't be a problem).
    remote_time = (year, month, monthday,
                   hour, min, sec,
                   -1, -1, local_time[-1])

    # Compute bogus seconds-since-epoch, and correct for difference
    # between local and remote timezones.
    bogus_rtime = mktime(remote_time)
    rtime = bogus_rtime - (remote_tz - local_tz)

    # And compute the difference between the "Date" header and
    # the current time.
    time_diff = now - rtime

    print "current time: %s UTC" % (asctime(gmtime(now)))
    print " remote time: %s UTC" % (asctime(gmtime(rtime)))
    print "difference = %.1f min" % (time_diff/60)

    # "Date" header is more than 24 hours ahead of current time
    # -- reject.
    if time_diff < - (MESSAGE_TOO_NEW * 3600):
        raise RejectMessage(
            "date header more than 24 hours ahead of current time\n"
            "(please set your computer's clock correctly)")

    # "Date" header is more than seven days behind current time --
    # reject.
    elif time_diff > (MESSAGE_TOO_OLD * 3600):
        raise RejectMessage(
            "date header more than 7 days behind current time\n"
            "(either your computer's clock is broken, or the Internet\n"
            "is really slow this week)")

    # Fall through to end -- message is accepted.
