# elspy library -- simplistic detection of "evil" attachments
#
# This is based on the "Generic Windows Executable Content" filter
# developed and no-longer-maintained by Nigel Metheringham, available
# on the Exim web site: http://www.exim.org/system_filter.exim .
# An updated version of this filter is also available on my
# web page at http://python.net/~gward/exim/ .
#
# Implementing this in Python with elspy has a number of benefits
# over detecting executable content in the system filter:
#   * since the message is rejected at SMTP time, you don't have
#     to worry about notifying the sender -- responsibility for
#     the message is foisted onto the remote SMTP client
#   * since you're using Exim's built-in rejection mechanism, the
#     message headers are logged like any other rejected message
#   * since it's written in Python, it's not constrained by
#     the limitations of Exim's filter language -- for one thing,
#     the pattern-matching is (IMHO) easier to read
#
# This is a simplified version that only blocks mail based on
# the names of attached files -- no special code for MyParty,
# SirCam, or the "sexyfun.net" virus.  (It works just fine
# without that special code.)

# XXX to do this "right" ie. with MIME-parsing, steal Neil's code:
#  http://arctrix.com/~nas/misc/qmail-filter-exe.py

# Copyright (c) 2002 Gregory P. Ward.  All rights reserved.
# See README.txt for licensing information.
#
# The elspy home page is at http://elspy.sourceforge.net/ .

__revision__ = "$Id: execontent_simple.py,v 1.11 2002/10/21 22:20:14 gward Exp $"

import os, re
from elspy import RejectMessage, get_fdpos, set_fdpos


BAD_EXTENSIONS = ("vbs", "vbe", "wsf", "wsh", "js", "jse",
                  "exe", "com", "cmd", "shs", "hta",
                  "bat", "scr", "pif", "lnk")

SIZE_LIMIT = 32*1024


fname_pat = (r'(\"(?P<fn1>[^\"]+       \. (?P<ext1>\w+))\"'
             r'|  (?P<fn2>[\x20-\xFF]+ \. (?P<ext2>\w+)))')


flags = re.VERBOSE | re.IGNORECASE

# Used for matching the "Content-type" header -- this is for
# messages whose whole content is executable.
msgname_re = re.compile(r'(?:file)?name=(%s)' % fname_pat, flags)

# Match "Content-Type: */*; " or "Content-Disposition: attachment; "
# -- will be followed (in attachname_re_{1,2} below) by
# "filename=..." or "name=...".
mimehdr_pat = (r'content-(?:type:\s*[\w-]+/[\w-]+|'
               r'           disposition:\s*attachment)'
               r'\s*;\s*')

# Used for searching the entire message body -- the idea is to look
# for a MIME "Content-Type" header somewhere in the message,
# on the assumption that it signifies an attachment.
attachname_re = re.compile(r'^%s (?:file)?name=(%s)'
                           % (mimehdr_pat, fname_pat),
                           flags|re.MULTILINE)

uuhdr_pat = r'^begin\s+[0-7]{3,4} \s+'
uuname_re = re.compile(r'%s (%s)' % (uuhdr_pat, fname_pat),
                       flags|re.MULTILINE)


def check_filename (match, what):
    if match is None:
        return
    fn = match.group('fn1') or match.group('fn2')
    ext = match.group('ext1') or match.group('ext2')
    if ext.lower() in BAD_EXTENSIONS:
        raise RejectMessage(
            "message rejected -- looks like a virus\n"
            "(name of %s (%r) indicates executable content)"
            % (what, fn))

def local_scan (info, headers, fd):

    # First check the message's "Content-type" header, to handle the
    # case of a message whose entire body is a Windows executable.
    ctype = headers.get("Content-type")
    if ctype:
        match = msgname_re.search(ctype)
        check_filename(match, "message body")
    
    # Read in the first chunk (32k by default) of the message body, and
    # scan it for evil attachments or uuencoded files.
    fpos = get_fdpos(fd)
    body = os.read(fd, SIZE_LIMIT)
    set_fdpos(fd, fpos)

    pos = 0
    while 1:
        match = attachname_re.search(body, pos)
        if match:
            check_filename(match, "attachment")
            pos = match.end()       # look for next attachment
        else:
            break

    match = uuname_re.search(body)
    check_filename(match, "uuencoded file")


if __name__ == "__main__":
    fname_re = re.compile(fname_pat, flags)
    fn1 = 'foo  .baz.ex .COM'
    fn2 = '"ding-do+n g.scr"'

    match = fname_re.match(fn1)
    assert match
    fn = match.group('fn1') or match.group('fn2')
    ext = match.group('ext1') or match.group('ext2')
    assert fn == fn1
    assert ext == "COM" and ext.lower() in BAD_EXTENSIONS

    match = fname_re.match(fn2)
    assert match
    fn = match.group('fn1') or match.group('fn2')
    ext = match.group('ext1') or match.group('ext2')
    assert fn == fn2[1:-1], "%r != %r" % (fn, fn2[1:-1])
    assert ext == "scr" and ext.lower() in BAD_EXTENSIONS

    
    
