"""elspy.maildir

Provides the Maildir class, a simple write-only Maildir implementation.
Handy for saving messages from local_scan() without bothering Exim (or
changing the recipient list, or relying on some funky special-purpose
router, or anything like that).
"""

# Copyright (c) 2002 Gregory P. Ward.  All rights reserved.
# See README.txt for licensing information.
#
# The elspy home page is at http://elspy.sourceforge.net/ .

__revision__ = "$Id: maildir.py,v 1.4 2002/10/21 22:20:14 gward Exp $"

import os
import socket
import time
import errno
from elspy.util import transfer_data

def _repr (self):
    return "<%s at %x: %s>" % (self.__class__.__name__, id(self), self)


class Maildir:
    """
    Simple write-only Maildir implementation.  For info on the Maildir
    format, see
      http://cr.yp.to/proto/maildir.html

    This lets you save messages to a folder from your local_scan()
    function completely independently of Exim.

    Sample usage:

      mdir = Maildir("/var/mail/spam", dir_mode=0770, file_mode=0640)
      msgfile = mdir.open_message()
      try:
          msgfile.write(...)
          mdir.close_message(msgfile)
      finally:
          msgfile.cleanup()

    The data passed to write() should be headers, newline, body, eg.

      f.write("From: joe@example.com\n")
      f.write("To: tim@example.net\n")
      f.write("\n")
      f.write("Hi Tim!\n")

    Of course you can combine all those in one call to write().

    For convenience, you can write a message in one go, as long as
    you have it in the form that Exim/elspy pass it to you in:

      mdir.add_message(headers, body_fd)

    where 'headers' is an instance of elspy.util.HeaderList, and
    'body_fd' is a readable file descriptor pointing to the start of the
    message body.
    """

    def __init__ (self, dir, dir_mode=0700, file_mode=0600):
        self.dir = dir
        self.dir_mode = dir_mode
        self.file_mode = file_mode

        self._ensure_exists()

    def __str__ (self):
        return self.dir

    __repr__ = _repr

    def join (self, *names):
        return os.path.join(self.dir, *names)

    def _ensure_exists (self):
        """
        Ensure that self.dir is really a Maildir, ie. is a directory
        with "tmp", "new", and "cur" subdirectories.  If not, create
        whatever directories are missing to make it a Maildir.
        """
        if not os.path.isdir(self.dir):
            os.mkdir(self.dir, self.dir_mode)
        for d in ("tmp", "new", "cur"):
            dir = self.join(d)
            if not os.path.isdir(dir):
                os.mkdir(dir, self.dir_mode)

    def _gen_filename (self):
        hostname = socket.gethostname()
        pid = os.getpid()
        num_tries = 0
        max_tries = 5
        while 1:
            basename = "%d.%d.%s" % (time.time(), pid, hostname)
            filename = self.join("tmp", basename)

            ok = 0                          # assume the worst
            num_tries += 1
            try:
                os.stat(filename)
            except OSError, err:
                # Good: file called 'filename' doesn't already exist.
                if err.errno == errno.ENOENT:
                    ok = 1

            if ok:
                break
            else:
                if num_tries > max_tries:
                    raise RuntimeError(
                        "could not create temporary file in %s/tmp"
                        % self.dir)
                time.sleep(2)           # and try again

        return (basename, filename)

    def open_message (self, basename=None):
        """open_message() -> MessageFile

        Open a new message file in the "tmp" subdirectory of self.dir.
        Returns a MessageFile object, which provides a file-like
        interface to the new message file (ie. you can call write() on
        it).
        """
        if basename is None:
            (basename, filename) = self._gen_filename()
        else:
            filename = self.join("tmp", basename)
            if os.path.exists(filename):
                raise RuntimeError(
                    "file %s already exists in %s"
                    % (`basename`, self.join("tmp")))

        flags = os.O_WRONLY|os.O_EXCL|os.O_CREAT
        fd = os.open(filename, flags, self.file_mode)
        return MessageFile(self, basename, fd)

    def close_message (self, msgfile, dest="new", status=None):
        """close_message(msgfile : MessageFile)

        Finish writing a message to the Maildir: closes the file
        descriptor and moves the message file from 'tmp' to 'cur'
        in the Maildir.
        """
        # Close the message file
        msgfile.flush()
        msgfile.file.close()
        
        # Start renaming the message file: link $dest/$basename name to
        # tmp/$basename.
        tmp_name = self.join("tmp", msgfile.basename)
        new_name = self.join(dest, msgfile.basename)
        if status:
            new_name += ":" + status
        os.link(tmp_name, new_name)

        # It's the responsibility of the caller to finish the rename by
        # calling msgfile.cleanup(), which just deletes the tmp name.
        # (This should be done in a "finally" clause, so it happens if
        # *anything* goes wrong while writing the message or renaming
        # the tmp file.)


    def write_message (self, headers, body_fd, msgfile):
        """
        Write a complete message to a message file -- headers, then
        a newline, then body.
        """
        headers.write(msgfile)
        msgfile.write("\n")
        msgfile.flush()
        transfer_data(body_fd, msgfile.fileno())

    def add_message (self, headers, body_fd):
        """add_message(headers : elspy.util.HeaderList,
                       body_fd : file)

        Add a message to the Maildir.  The message headers are in
        'headers', and the body is ready to be read from 'body_fd'.
        """
        # Open the file in tmp.
        msgfile = self.open_message()

        try:
            # Write the message to it (headers, newline, body) and close
            # the message file.
            self.write_message(msgfile, headers, body_fd)
            self.close_message(msgfile)
        finally:
            # Remove the temporary name of the message file.
            msgfile.cleanup()

# class Maildir


class MessageFile:
    """
    A single, just-created message in a Maildir, ready to have the
    message headers and body written to it.  MessageFile objects should
    only be created by Maildir.open_message(); they provide almost all
    file methods, so you can call write() or flush() directly on them.

    The one file method not provided is close() -- to close a
    MessageFile, use the close_message() method on the Maildir that
    created it.
    """

    def __init__ (self, maildir, basename, fd):
        self.maildir = maildir
        self.basename = basename
        self.file = file = os.fdopen(fd, "w")

        # self.write = self.file.write, etc.
        for name in dir(file):
            if name[0] == "_" or name == "close":
                continue
            obj = getattr(file, name)
            if callable(obj):
                setattr(self, name, obj)

    def __str__ (self):
        return self.basename or "*no basename*"

    __repr__ = _repr

    def cleanup (self):
        """
        Remove the temporary file used to write the message.  You should
        call this in a "finally" clause after writing the message and
        closing the message file (see example in Maildir class
        docstring).
        """
        if self.basename is None:       # already cleaned up
            return

        fullname = self.maildir.join("tmp", self.basename)
        os.unlink(fullname)
        self.basename = self.maildir = None


if __name__ == "__main__":
    md = Maildir("mdir")
    mf = md.open_message(basename="123-abc")
    mf.write("Header1: blah\nHeader2: dinkum\n  dankum\n\n")
    mf.write("body here\n")
    md.close_message(mf, dest="cur", status="2,S")
    mf.cleanup()
    
