# elspy library -- save mid-range spam for certain users in a holding pen
#
# This local_scan() is for use in conjunction with the elspy
# SpamAssassin front-end in spamassassin.py -- you must call this one
# *after* elspy.spamassassin.local_scan(), since it relies on the
# presence of SpamAssassin headers.
# 
# Since the elspy SA front-end only rejects messages with a fairly high
# spam score, some mail that SA considers spam will still slip through.
# In particular, any message that scores between 5 and 10 (the default
# thresholds used by SpamAssassin and the elspy SA front-end) is
# considered "mid-range" spam.  If all you do is use elspy.spamassassin
# to reject high-scoring spam, then mid-range spam will still find its
# way to users' inboxes.  This may or may not be desirable at your site;
# this module implements one way to deal with the problem.
# 
# In particular, mid-range spam destined for addresses in the
# ROUTESPAM_ADDRS list will be saved to the Maildir folder named by
# SPAM_FOLDER.  Those addresses will be removed from the message's
# recipient list, but added to an "Envelope-To" header in the copy saved
# to SPAM_FOLDER.  If all of the message's recipients are in
# ROUTESPAM_ADDRS, then the message will be blackholed by Exim -- the
# only surviving copy will be in SPAM_FOLDER.
# 
# If you think it's Wrong and Evil for local_scan() code to be saving
# messages to a Maildir at SMTP-time, then don't use this code.  There's
# more than one way to do it.

# Copyright (c) 2002 Gregory P. Ward.  All rights reserved.
# See README.txt for licensing information.
#
# The elspy home page is at http://elspy.sourceforge.net/ .

__revision__ = "$Id: routespam.py,v 1.10 2002/10/21 22:20:14 gward Exp $"

import os, socket, time, errno
from elspy import get_fdpos, set_fdpos, log_write, LOG_REJECT
from elspy.maildir import Maildir

ROUTESPAM_ADDRS = []

SPAM_FOLDER = "/var/mail/spam"

SPAM_FOLDER_UMASK = 077


def save_spam (folder, body_fd, headers, recips, sender):
    mdir = Maildir(folder,
                   dir_mode=0777 & ~SPAM_FOLDER_UMASK,
                   file_mode=0666 & ~SPAM_FOLDER_UMASK)
    msgfile = mdir.open_message()
    pos = get_fdpos(body_fd)
    try:
        msgfile.write("Return-Path: <%s>\n" % sender)
        msgfile.write("Envelope-To: %s\n" % ", ".join(recips))
        mdir.write_message(headers, body_fd, msgfile)
        mdir.close_message(msgfile)
    finally:
        msgfile.cleanup()
        set_fdpos(body_fd, pos)

def local_scan (info, headers, fd):
    routespam_addr = {}
    for addr in ROUTESPAM_ADDRS:
        routespam_addr[addr] = 1

    if headers.get("x-spam-flag", "").strip().upper() != "YES":
        return

    i = 0
    remove_idx = []                 # list of list indeces
    for addr in info.recipients_list:
        if routespam_addr.has_key(addr) or routespam_addr.has_key("*"):
            remove_idx.append(i)
        i += 1

    if remove_idx:
        remove_addrs = [info.recipients_list[i] for i in remove_idx]
        save_spam(SPAM_FOLDER, fd, headers,
                  remove_addrs, info.sender_address)
        remove_idx.reverse()
        for i in remove_idx:
            del info.recipients_list[i]

        remove_addrs = ", ".join(remove_addrs)
        if info.recipients_list:
            msg = ("saved to %s for %s (accepted for %s)"
                   % (SPAM_FOLDER, remove_addrs,
                      ", ".join(info.recipients_list)))
        else:
            msg = ("saved to %s for %s (no other recipients)"
                   % (SPAM_FOLDER, remove_addrs))
        log_write(LOG_REJECT, msg)

    # Fall off the end -- either accept the message for the remaining
    # recipients, or blackhole it if no recipients remain.
