# elspy library -- pass a message through SpamAssassin
#
# This local_scan() emulates the spamc client included with
# SpamAssassin.  We don't use spamc itself because we don't need to --
# all that's needed is to open a socket to spamd, feed it the message
# headers and body, and read the results.  This saves the overhead
# of spawning spamc.
# 
# If the score returned by spamd is high enough (configured by setting
# REJECT_THRESHOLD below), the message is rejected.  Otherwise, we tack
# on a couple of headers so that other local agents can use the results
# of SpamAssassin's analysis.
# 
# Specifically, we add the following headers to all accepted messages,
# regardless of their spam score:
#   X-Spam-Status
#   X-Spam-Level
#
# X-Spam-Status looks like this for a non-spam message (reformatted
# for readability):
#   X-Spam-Status: No, hits=3.4 required=5.0
#                  tests=FROM_NAME_NO_SPACES,DATE_MISSING,TO_MALFORMED,
#                        INVALID_DATE
# and for a spam message:
#   X-Spam-Status: Yes, hits=10.4 required=5.0
#                  tests=TO_MALFORMED,PLING_PLING,NO_REAL_NAME,INVALID_DATE,
#                        PLING,PORN_10,UPPERCASE_50_75,NORMAL_HTTP_TO_IP,
#                        PORN_3,SUBJ_ALL_CAPS
#
# X-Spam-Level is a simple regex-friendly way of representing the spam
# score, eg.
#   X-Spam-Level: ***         (3.0 <= score < 4.0)
#   X-Spam-Level: **********  (10.0 <= score < 11.0)
#
# Furthermore, if spamd considers the message spammy, but not spammy
# enough to be rejected, we add
#   X-Spam-Flag: YES
# 
# (Unfortunately, we don't add "X-Spam-Report" because there doesn't
# appear to be an easy way to get both the list of tests and
# the spam report out of spamd.  Sigh.)

# Copyright (c) 2002 Gregory P. Ward.  All rights reserved.
# See README.txt for licensing information.
#
# The elspy home page is at http://elspy.sourceforge.net/ .

__revision__ = "$Id: spamassassin.py,v 1.15 2002/10/21 22:20:14 gward Exp $"

import os, re, errno
from socket import socket, AF_INET, SOCK_STREAM, error as SocketError
from elspy import log_warning, log_error, get_fdpos, set_fdpos, \
     RejectMessage, TempRejectMessage


# ----------------------------------------------------------------------
# Configuration

# Where to connect to spamd.
SPAMD_HOST = "127.0.0.1"
SPAMD_PORT = 783

# Maximum message size (bytes) that will be passed to spamd.
MAX_SIZE = 250*1024

# Which user spamd should become in order to process messages.
SPAMD_USER = "mail"

# If the message scores >= REJECT_THRESHOLD, we reject it outright.
REJECT_THRESHOLD = 10.0

# -- end of configuration ----------------------------------------------

# Raised if we have problems connecting to or sending data to spamd.
class ConnectionError (Exception):
    def __init__ (self, errno, errmsg):
        self.errno = errno
        self.errmsg = errmsg

    def __str__ (self):
        return self.errmsg

# Raised if we can't understand what spamd is saying to us.
class ProtocolError (Exception):
    pass


def open_connection ():
    try:
        sock = socket(AF_INET, SOCK_STREAM)
        sock.connect((SPAMD_HOST, SPAMD_PORT))
    except SocketError, err:
        raise ConnectionError, err.args

    return sock

def send_message (sock, fd, headers):

    # Send request to spamd.  "Content-length" is optional so skip it.
    # "User" should be supplied in case spamd is running as root -- we
    # don't want to force it to fallback to "nobody" every time, since
    # that could fail, and it's just more noise in the logs.
    req = ("SYMBOLS SPAMC/1.2\r\n"
           "User: %s\r\n"
           "\r\n" % SPAMD_USER)
    sfile = sock.makefile("w", 0)
    sfile.write(req)

    # Write headers to spamd socket.
    try:
        headers.write(sfile)
        sfile.write("\n")
    except IOError, err:
        raise ConnectionError, err.args

    # And write the body.
    fpos = get_fdpos(fd)
    while 1:
        data = os.read(fd, 1024)
        if not data:
            break
        try:
            sfile.write(data)
        except IOError, err:
            raise ConnectionError, err.args
    set_fdpos(fd, fpos)
    sfile.flush()

    # Shutdown the sending end of the socket, so spamd knows
    # to process the message now.
    try:
        sock.shutdown(1)
    except SocketError, err:
        raise ConnectionError, err.args


status_line_re = re.compile(r'Spam: \s+ (True|False)'
                            r'\s* ; \s*'
                            r'([\.\d\-]+) \s* / \s* ([\.\d\-]+)',
                            re.VERBOSE)

def read_response (sock):

    sfile = sock.makefile("r")
    protocol_line = sfile.readline()
    if not protocol_line:
        raise ConnectionError(None, "spamd closed connection early")

    try:
        (protocol, status, message) = protocol_line.split()
    except ValueError:
        raise ProtocolError("unable to parse first response line: %r"
                            % protocol_line)
    if int(status) != 0 or message != "EX_OK":
        raise ProtocolError("error processing message: %s" % message)

    status_line = sfile.readline()
    match = status_line_re.match(status_line)
    if not match:
        raise ProtocolError("couldn't parse spamd output: %r" % status_line)
    is_spam = match.group(1) == "True"
    score = float(match.group(2))
    threshold = float(match.group(3))

    blank = sfile.readline()
    if blank.strip():
        raise ProtocolError("expected blank line from spamd, got %r" % blank)

    tests_line = sfile.readline()
    tests = tests_line.strip().split(",")

    empty = sfile.readline()
    if empty:
        raise ProtocolError("expected end of output from spamd, got %r"
                            % empty)

    return (is_spam, score, threshold, tests)


def local_scan (info, headers, fd):

    # Header names for easy typing
    xss = "X-Spam-Status"
    xsf = "X-Spam-Flag"
    xsl = "X-Spam-Level"

    pos = get_fdpos(fd)
    msg_size = os.fstat(fd)[6] - pos
    if msg_size > MAX_SIZE:
        why = "%d bytes, max %d" % (msg_size, MAX_SIZE)
        log_warning("message too big for SpamAssassin (%s) -- not checking"
                    % why)
        headers.add(xss, "not checked (message too big: %s)" % why)
        headers.add(xsf, "UNKNOWN")
        return

    try:
        sock = open_connection()
        send_message(sock, fd, headers)
        (is_spam, score, threshold, tests) = read_response(sock)
    except ConnectionError, err:
        # XXX should retry a few times before giving up
        log_warning("spamd not available (%s) -- accepting message" % err)
        return

    # Score is over the top -- reject the message outright.
    if score >= REJECT_THRESHOLD:
        msg = ("rejecting message -- looks like spam\n"
               "(score: %.1f; reasons: %s)"
               % (score, ",".join(tests)))
        raise RejectMessage(msg)

    # Headers to add to all accepted messages (whether SA considers
    # them spam or not).
    spam_status = ("%s, hits=%.1f required=%.1f tests=%s"
                   % (is_spam and "Yes" or "No",
                      score, threshold, ",".join(tests)))
    spam_level = "*" * int(score)

    # SpamAssassin thinks this is spam, but it's below the reject
    # threshold.  Add "this is spam" headers and accept the message.
    if is_spam:
        headers.add(xss, spam_status)
        headers.add(xsf, "YES")
        headers.add(xsl, spam_level)

    # Probably not spam -- don't add X-Spam-Flag.
    else:
        headers.add(xss, spam_status)
        headers.add(xsl, spam_level)

    # Fallthrough -- accept the message
