"""elspy.util

Various utility code useful to site-specific local_scan() functions,
and needed by the default _local_scan() in elspy/__init__.py.
"""

# Copyright (c) 2002 Gregory P. Ward.  All rights reserved.
# See README.txt for licensing information.
#
# The elspy home page is at http://elspy.sourceforge.net/ .

__revision__ = "$Id: util.py,v 1.9 2002/10/21 22:20:14 gward Exp $"

import os
from types import StringType
from elspy import \
     log_write, header_add, LOG_MAIN, \
     LOCAL_SCAN_ACCEPT, LOCAL_SCAN_TEMPREJECT, LOCAL_SCAN_REJECT


class HeaderList:

    def __init__ (self, headers):
        self._headers = []
        for (_, text) in headers:
            self._addline(text)

    def _addline (self, text):
        self._headers.append(tuple(text.split(":", 1)))

    def __str__ (self):
        num = len(self._headers)
        if num > 5:
            return ("%d headers: %s, %s, ..., %s"
                    % (num, self._headers[0][0],
                            self._headers[1][0],
                            self._headers[-1][0]))
        else:
            return ("%d headers: %s"
                    % (num, ", ".join([n for (n, v) in self._headers])))

    def __repr__ (self):
        return "<%s at 0x%x: %s>" % (self.__class__.__name__, id(self), self)

    def __len__ (self):
        return len(self._headers)

    def __getitem__ (self, i):
        return self._headers[i]

    def has_header (self, name):
        return self.get_first(name) is not None

    def get_first (self, name, default=None):
        name = name.lower()
        for (n, v) in self._headers:
            if n.lower() == name:
                return v
        else:
            return default

    get = get_first

    def get_last (self, name, default=None):
        name = name.lower()
        for i in range(len(self._headers)-1, -1, -1):
            (n, v) = self._headers[i]
            if n.lower() == name:
                return v
        else:
            return default

    def get_all (self, name):
        name = name.lower()
        values = []
        for (n, v) in self._headers:
            if n.lower == name:
                values.append(v)
        return values

    def add (self, name, value):
        sp = nl = ""
        if value and value[0] not in " \t":
            sp = " "
        if not value or value[-1] != "\n":
            nl = "\n"
        value = sp + value + nl
        header_add("%s:%s" % (name, value))
        self._headers.append((name, value))

    def as_string (self):
        return "".join(["%s:%s" % nameval for nameval in self._headers])

    def write (self, file=None):
        if file is None:
            file = sys.stdout
        for nameval in self._headers:
            file.write("%s:%s" % nameval)


class EximInfo:
    def __init__ (self, dict):
        self.__dict__ = dict

    def __repr__ (self):
        return "<%s at 0x%x>" % (self.__class__.__name__, id(self))


class MessageDisposition (Exception):
    def __init__ (self, text=None):
        # do not allow Unicode, because SMTP doesn't speak Unicode
        if text is None or isinstance(text, StringType):
            self.text = text
        else:
            raise TypeError("'text' must be a string or None")

    def __str__ (self):
        return self.text

class AcceptMessage (MessageDisposition):
    disposition = LOCAL_SCAN_ACCEPT

class TempRejectMessage (MessageDisposition):
    disposition = LOCAL_SCAN_TEMPREJECT

class RejectMessage (MessageDisposition):
    disposition = LOCAL_SCAN_REJECT


def log_error (msg):
    log_write(LOG_MAIN, "error: " + msg)

def log_warning (msg):
    log_write(LOG_MAIN, "warning: " + msg)


def get_fdpos (fd):
    """Return the current file offset for OS-level file descriptor 'fd'."""
    return os.lseek(fd, 0, 1)

def set_fdpos (fd, pos):
    """Set the file offset for OS-level file descriptor 'fd'."""
    os.lseek(fd, pos, 0)
    

def transfer_data (in_fd, out_fd):
    """Read data from in_fd and write it to out_fd.
    """
    chunk = 16*1024
    while 1:
        data = os.read(in_fd, chunk)
        if not data:
            break
        n = os.write(out_fd, data)
        while n < len(data):
            n += os.write(out_fd, data[n:])
    os.fsync(out_fd)
