"""Common code for the elspy testing scripts.
"""

import sys, os, re
import tempfile
import new
from rfc822 import Message

import elspy
from elspy.util import HeaderList, EximInfo, \
     get_fdpos, set_fdpos, transfer_data


__revision__ = "$Id: test_util.py,v 1.2 2002/10/20 21:49:03 gward Exp $"


def get_module (name):
    """(name : string) -> module

    If 'name' is a filename, execfile() it and return an artificial
    module object.

    If 'name' is a module name, import it and return the module object.
    """
    # If test_module is a filename, then execfile() it.
    if os.path.isfile(name):
        mod_name = os.path.splitext(os.path.basename(name))[0]
        mod = new.module(mod_name)
        mod_dict = mod.__dict__
        execfile(name, mod_dict, mod_dict)
        return mod

    # Otherwise assume it's a module name, and import it.
    else:
        __import__(name)
        return sys.modules[name]

def get_local_scan (mod, name='local_scan'):
    try:
        return getattr(mod, name)
    except AttributeError:
        sys.exit("test module %r has no %s() function" % (mod, name))
    

def create_header_list (msg):
    headers = []
    htype = ord(' ')
    prev_line = None
    for line in msg.headers:
        if line[0] in ' \t':
            prev_line += line
        else:
            if prev_line is not None:
                headers.append((htype, prev_line))
            prev_line = line
    headers.append((htype, prev_line))

    return headers

def create_body_file (msg_file):
    filename = tempfile.mktemp()
    fd = os.open(filename, os.O_CREAT|os.O_RDWR)
    os.write(fd, "000000-000000-00\n")
    pos = get_fdpos(fd)
    msg_fd = msg_file.fileno()
    set_fdpos(msg_fd, msg_file.tell())
    transfer_data(msg_fd, fd)
    set_fdpos(fd, pos)

    os.unlink(filename)
    #print "body fd=%d; pos=%d" % (fd, pos)
    #time.sleep(30)
    return fd

def create_info_object (msg):
    """(msg : rfc822.Message) -> info : EximInfo
    """

    # Create an empty Exim info object.
    info = EximInfo({"sender_address" : None,
                     "interface_address" : None,
                     "interface_port" : None,
                     "received_protocol" : None,
                     "recipients_list" : [],
                     "sender_host_address" : None,
                     "sender_host_authenticated" : None,
                     "sender_host_name" : None,
                     "sender_host_port" : None,
                     })
                    
    # And fill in what we can from the headers.
    sender = msg.get('return-path') or msg.get('envelope-from')
    if not sender and msg.unixfrom:
        sender = msg.unixfrom.split()[0]
    if sender:
        info.sender_address = sender

    recips = msg.get('envelope-to')
    if recips:
        recips = re.split(r',\s*', recips)
        bracket_re = re.compile(r'<([^>]+)>')
        recips = [bracket_re.sub('\1', recip) for recip in recips]
        info.recipients_list = recips

    # Could probably parse the first "Received" header and fill in
    # a lot of the other fields in 'info', but why bother?
    

def create_message (filename):
    """(filename : string) -> (body_fd : int,
                               headers : HeaderList,
                               info : EximInfo)

    Read an RFC 2822 message from 'filename', and create the
    (info, headers, fd) triple that elspy normally passes to local_scan().
    """

    # Read headers from the message.
    msg_file = open(filename, "r")
    msg = Message(msg_file)

    # Convert to the Python analog of Exim's internal format, and thence
    # to elspy's HeaderList class.
    headers = create_header_list(msg)
    headers = HeaderList(headers)

    # Copy message body to temporary file that looks like an Exim
    # spool file.
    fd = create_body_file(msg_file)

    # Create an Exim info object, and try to fill in a few fields
    # from the message headers.
    info = create_info_object(msg)

    return (info, headers, fd)

def interpret_disposition (exc, max_text=None):
    """(exc : MessageDisposition) -> (what : string, text : string)
    """
    what = { elspy.LOCAL_SCAN_ACCEPT: "accepted",
             elspy.LOCAL_SCAN_REJECT: "rejected",
             elspy.LOCAL_SCAN_TEMPREJECT: "temporarily rejected"
           } [exc.disposition]
    text = massage_text(exc.text, max_text)
    return (what, text)

def massage_text (text, max_len=None):
    if text is None:
        return "(no return text)"
    else:
        text = `text`
        if max_len is not None and len(text) > max_len:
            text = text[0:max_len-5] + " ..." + text[0]
        return text
